import random, math

class BasicRandomSolver(object):
    name ="Basic Random Solver"

    def __init__(self, board, k):
        self.board = board
        self.k = k

    def choose_column(self):
        v = self.board.v
        counters = [[0,i] for i in range(self.board.nx)]
        for x in range(self.board.nx):
            if self.board.put(x):
                if self.board.won == v: #ia wins
                    self.board.unput(1)
                    return x #for coin put at line 13
                else:
                    l1 = len(self.board.history)
                    counters[x][0] = self.random_plays(v)
                    assert l1 == len(self.board.history)
                    self.board.unput(1) #for coin put at line 13
            else:
                counters[x][0] = -float("inf")
##        print("counters", counters)
        return max(counters)[1]


    def random_plays(self,v):
        i = 0
        counter = 0
        hlength = len(self.board.history)
        while i < self.k:
            if len(self.board.history) < len(self.board):
                x = random.randint(0,self.board.nx-1)
                if self.board.put(x):
                    if self.board.won is not None: #if game is finished
                        i += 1 #increment number of games
                        if self.board.won == v:
                            counter += 1
                        else:
                            counter -= 1
                        self.board.unput(len(self.board.history)-hlength)
            else:
                i += 1
                self.board.unput(len(self.board.history)-hlength)
        assert hlength == len(self.board.history)
        return counter

class SmallSolution(object):

    @staticmethod
    def get_max(sols):
        best_val = -float("inf")
        best_sol = None
        for s in sols:
            if s.score > best_val:
                best_val = s.score
                best_sol = s
        return best_sol

    @staticmethod
    def get_min(sols):
        best_val = float("inf")
        best_sol = None
        for s in sols:
            if s.score < best_val:
                best_val = s.score
                best_sol = s
        return best_sol


    def __init__(self, column, score):
        self.column = column
        self.score = score

class ImprovedRandomSolver(BasicRandomSolver):
    name = "Improved Random Solver"

    def get_children(self):
        children = []
        for x in range(self.board.nx):
            if self.board.put(x):
                won = self.board.won
                self.board.unput(1)
                if won:
                    return [x]
                else:
                    children.append(x)
        return children

    def choose_column(self):
        children = self.get_children()
        if len(children) == 1:
            return children[0]
        solutions = [SmallSolution(column, 0) for column in children]
        for s in solutions:
            self.board.put(s.column)
            s.score = self.random_plays()
            self.board.unput(1)
        if self.board.v < 0:
            return SmallSolution.get_min(solutions).column
        else:
            return SmallSolution.get_max(solutions).column

    def random_play(self):
        hlength = len(self.board.history)
        while len(self.board.history) < len(self.board): #while not filled up
            children = self.get_children()
            x = random.choice(children)
            self.board.put(x)
            won = self.board.won
            if won:
                self.board.unput(len(self.board.history)-hlength)
                return won
        self.board.unput(len(self.board.history)-hlength)
        return 0

    def random_plays(self):
        counter = 0
        for i in range(self.k):
            counter += self.random_play()
        return counter


class MinMaxSolver(object):
    name = "MinMax Solver"

    def __init__(self, board, max_depth):
        self.board = board
        self.max_depth = max_depth
        #k/(depth+1) > depth ==> k > depth(depth+1)
        self.k = 1e6
        self.cx = self.board.nx // 2
        self.cy = self.board.ny // 2
        self.evaluation = self.evalution2

    def evalution1(self, depth):
        self.board.v * depth #so that try to live as long as possible

    def get_distance(self, x, y):
        return abs(self.cx-x) + abs(self.cy-y) #manhattan
##        return math.hypot(self.cx-x, self.cy-y)

    def evalution2(self, depth):
        value = 0
        for x,y in self.board:
            distance = self.get_distance(x,y)
            value += self.board[x,y]/(distance+1)
        return value

    def solve(self, depth):
        if len(self.board.history) < len(self.board):
            children = {}
            for x in range(self.board.nx):
                if self.board.put(x):
                    won = self.board.won
                    #handle win
                    if won: #happens necessarily with won = seld.board.v (=min for v=-1, =max for v=1)
                        self.board.unput(1)
##                        assert won == self.board.v #debug ok
                        return x, won*self.k/(depth+1)
                    #collect result
                    if depth < self.max_depth:
                        col, result = self.solve(depth+1)
                    else:
                        result = self.evaluation(depth)
                    #add result
                    if result in children:
                        children[result] += [x]
                    else:
                        children[result] = [x]
                    #unput
                    self.board.unput(1)
            if self.board.v < 0:
                best_result = min(children)
            elif self.board.v > 0:
                best_result = max(children)
            else:
                raise Exception("problem")
            if depth == 0:
                print(children, best_result)
            randx = random.choice(children[best_result])
            return randx, best_result
        else:
            return -1, 0

#optimisation : fonction speciale pour iterer seulement sur colonnes accessibles


    def choose_column(self):
        column, score = self.solve(0)
        print("chosen", score, column)
        return column


class CompositeSolver(object):
    name = "Composite Solver"

    def __init__(self, board, max_depth, k, transition_n):
        self.board = board
        self.transition_n = transition_n
        self.random_solver = BasicRandomSolver(board, k)
        self.minmax_solver = MinMaxSolver(board, max_depth)

    def choose_column(self):
        if len(self.board) - len(self.board.history) <= self.board.n:
            self.minmax_solver.max_depth = self.board.n
            print("max depth!")
        if len(self.board.history) > self.transition_n:
##            print("minmax")
            return self.minmax_solver.choose_column()
        else:
##            print("random")
            return self.random_solver.choose_column()